/******************************************************************************/
/*                                                                            */
/*                Advanced Navigation Packet Protocol Library                 */
/*                C Language Dynamic Orientus SDK, Version 6.0                */
/*                    Copyright 2022, Advanced Navigation                     */
/*                                                                            */
/******************************************************************************/
/*
 * Copyright (C) 2022 Advanced Navigation
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

#define RS232 0
#define NETWORK 1

#define CONNECTION_TYPE NETWORK

#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <math.h>
#include <time.h>
#include <string.h>

#ifdef _WIN32
#include <windows.h>
#if CONNECTION_TYPE==NETWORK
#define _WIN32_WINNT 0x0501
#include <winsock2.h>
#include <ws2tcpip.h>
#endif
#else
#include <unistd.h>
#if CONNECTION_TYPE==RS232
#include "rs232/rs232.h"
#elif CONNECTION_TYPE==NETWORK
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netdb.h>
#endif
#endif

#include "an_packet_protocol.h"
#include "ins_packets.h"

#define RADIANS_TO_DEGREES (180.0/M_PI)

static unsigned char request_all_configuration[] = { 0xE2, 0x01, 0x10, 0x9A, 0x73, 0xB6, 0xB4, 0xB5, 0xB8, 0xB9, 0xBA, 0xBC, 0xBD, 0xC0, 0xC2, 0xC3, 0xC4, 0x03, 0xC6, 0x45, 0xC7 };
int socket_fd = -1;

int transmit(unsigned char* data, int length);
int receive(unsigned char* data, int length);
int an_packet_transmit(an_packet_t* an_packet);
void set_filter_options();

int main(int argc, char* argv[])
{
	an_decoder_t an_decoder;
	an_packet_t* an_packet;
	
	system_state_packet_t system_state_packet;
	raw_sensors_packet_t raw_sensors_packet;
	
	FILE* anpp_log_file;
	
	char filename[64];
	time_t rawtime;
	struct tm* timeinfo;
	int write_counter = 0;
	int bytes_received;
	
#if CONNECTION_TYPE==RS232
	if(argc != 3)
	{
		
#ifdef _WIN32
		printf("Usage - program com_port baud_rate\nExample - packet_example.exe COM1 115200\n");
#else
		printf("Usage - program com_port baud_rate\nExample - packet_example.exe /dev/ttyUSB0 115200\n");
#endif	
		exit(EXIT_FAILURE);
	}
	
	/* open the com port */
	if(OpenComport(argv[1], atoi(argv[2])))
	{
		printf("Could not open serial port\n");
		exit(EXIT_FAILURE);
	}
	
#elif CONNECTION_TYPE==NETWORK
	if(argc != 3)
	{
		printf("Usage - program ip_address tcp_port\nExample - packet_example.exe 192.168.1.54 16718\n");
		exit(EXIT_FAILURE);
	}
	
#if _WIN32
	WSADATA wsa;
	if (WSAStartup(MAKEWORD(2,2),&wsa) != 0)
	{
		printf("Failed. Error Code : %d", WSAGetLastError());
		exit(EXIT_FAILURE);
	}
#endif

	struct addrinfo hints = {0}, *addresses;
	hints.ai_family = AF_INET;
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_protocol = IPPROTO_TCP;
	if(getaddrinfo(argv[1], argv[2], &hints, &addresses))
	{
		printf("Failure resolving hostname %s, port %s\n", argv[1], argv[2]);
		exit(EXIT_FAILURE);
	}

	if((socket_fd = socket(addresses->ai_family, addresses->ai_socktype, addresses->ai_protocol)) < 0)
	{
		printf("Socket creation error \n");
		freeaddrinfo(addresses);
		exit(EXIT_FAILURE);
	}

	if(connect(socket_fd, addresses->ai_addr, addresses->ai_addrlen) < 0)
	{
		printf("Socket connection failed \n");
		freeaddrinfo(addresses);
		exit(EXIT_FAILURE);
	}
	freeaddrinfo(addresses);
#else
	printf("Invalid CONNECTION_TYPE\n");
	exit(EXIT_FAILURE);
#endif
	
	/* Create a Log file */
	rawtime = time(NULL);
	timeinfo = localtime(&rawtime);
	sprintf(filename, "ANLog_%02d-%02d-%02d_%02d-%02d-%02d.anpp", timeinfo->tm_year - 100, timeinfo->tm_mon + 1, timeinfo->tm_mday, timeinfo->tm_hour, timeinfo->tm_min, timeinfo->tm_sec);
	anpp_log_file = fopen(filename, "wb");
	
	/* Request all the configuration and the device information from the unit */
	transmit(request_all_configuration, sizeof(request_all_configuration));
	
	an_decoder_initialise(&an_decoder);
	
	while(1)
	{
		if((bytes_received = receive(an_decoder_pointer(&an_decoder), an_decoder_size(&an_decoder))) > 0)
		{
			/* Log all binary data coming from the sensor, this can be converted to CSV using the manager software */
			fwrite(an_decoder_pointer(&an_decoder), sizeof(uint8_t), bytes_received, anpp_log_file);
			if(write_counter++ >= 100)
			{
				fflush(anpp_log_file);
				write_counter = 0;
			}
			
			/* increment the decode buffer length by the number of bytes received */
			an_decoder_increment(&an_decoder, bytes_received);
			
			/* decode all the packets in the buffer */
			while((an_packet = an_packet_decode(&an_decoder)) != NULL)
			{
				if(an_packet->id == packet_id_system_state) /* system state packet */
				{
					/* copy all the binary data into the typedef struct for the packet */
					/* this allows easy access to all the different values             */
					if(decode_system_state_packet(&system_state_packet, an_packet) == 0)
					{
						printf("System State Packet:\n");
						printf("\tLatitude = %f, Longitude = %f, Height = %f\n", system_state_packet.latitude * RADIANS_TO_DEGREES, system_state_packet.longitude * RADIANS_TO_DEGREES, system_state_packet.height);
						printf("\tRoll = %f, Pitch = %f, Heading = %f\n", system_state_packet.orientation[0] * RADIANS_TO_DEGREES, system_state_packet.orientation[1] * RADIANS_TO_DEGREES, system_state_packet.orientation[2] * RADIANS_TO_DEGREES);
					}
				}
				else if(an_packet->id == packet_id_raw_sensors) /* raw sensors packet */
				{
					/* copy all the binary data into the typedef struct for the packet */
					/* this allows easy access to all the different values             */
					if(decode_raw_sensors_packet(&raw_sensors_packet, an_packet) == 0)
					{
						printf("Raw Sensors Packet:\n");
						printf("\tAccelerometers X: %f Y: %f Z: %f\n", raw_sensors_packet.accelerometers[0], raw_sensors_packet.accelerometers[1], raw_sensors_packet.accelerometers[2]);
						printf("\tGyroscopes X: %f Y: %f Z: %f\n", raw_sensors_packet.gyroscopes[0] * RADIANS_TO_DEGREES, raw_sensors_packet.gyroscopes[1] * RADIANS_TO_DEGREES, raw_sensors_packet.gyroscopes[2] * RADIANS_TO_DEGREES);
					}
				}
				else
				{
					printf("Packet ID %u of Length %u\n", an_packet->id, an_packet->length);
				}
				
				/* Ensure that you free the an_packet when your done with it or you will leak memory */
				an_packet_free(&an_packet);
			}
		}
#ifdef _WIN32
    Sleep(10);
#else
	usleep(10000);
#endif
	}
}

int transmit(unsigned char* data, int length)
{
#if CONNECTION_TYPE==RS232
	return SendBuf(data, length);
#elif CONNECTION_TYPE==NETWORK
#if _WIN32
	return send(socket_fd, (char *) data, length, 0);
#else
	return write(socket_fd, data, length);
#endif
#endif
}

int receive(unsigned char* data, int length)
{
#if CONNECTION_TYPE==RS232
	return PollComport(data, length);
#elif CONNECTION_TYPE==NETWORK
#if _WIN32
	return recv(socket_fd, (char *) data, length, 0);
#else
	return read(socket_fd, data, length);
#endif
#endif
}

int an_packet_transmit(an_packet_t* an_packet)
{
	an_packet_encode(an_packet);
	return transmit(an_packet_pointer(an_packet), an_packet_size(an_packet));
}

/*
 * This is an example of sending a configuration packet to Orientus.
 *
 * 1. First declare the structure for the packet, in this case filter_options_packet_t.
 * 2. Set all the fields of the packet structure
 * 3. Encode the packet structure into an an_packet_t using the appropriate helper function
 * 4. Send the packet
 * 5. Free the packet
 */
void set_filter_options()
{
	an_packet_t* an_packet;
	filter_options_packet_t filter_options_packet;

	/* initialise the structure by setting all the fields to zero */
	memset(&filter_options_packet, 0, sizeof(filter_options_packet_t));

	filter_options_packet.permanent = TRUE;
	filter_options_packet.vehicle_type = vehicle_type_car;
	filter_options_packet.magnetometers_enabled = TRUE;

	an_packet = encode_filter_options_packet(&filter_options_packet);

	an_packet_transmit(an_packet);

	an_packet_free(&an_packet);
}
